@@ -10,6 +10,7 @@
 #define EVDI_DEBUG_H
 
 #include "evdi_params.h"
+#include <linux/version.h>
 
 #define EVDI_LOGLEVEL_FATAL   1
 #define EVDI_LOGLEVEL_ERROR   2
@@ -54,4 +55,10 @@
 
 void evdi_log_process(char *buf, size_t size);
 
+
+#if KERNEL_VERSION(6, 20, 0) <= LINUX_VERSION_CODE
+#else
+#define kzalloc_obj(obj, flags) kzalloc(sizeof(typeof(obj)), flags)
+#endif
+
 #endif /* EVDI_DEBUG_H */
