@@ -15,12 +15,15 @@
 #include <linux/dma-buf-map.h>
 #endif
 #if KERNEL_VERSION(5, 16, 0) <= LINUX_VERSION_CODE || defined(EL8) || defined(EL9)
+#include <drm/drm_gem_ttm_helper.h>
 #include <drm/drm_prime.h>
 #include <drm/drm_file.h>
+#include <linux/minmax.h>
 #elif KERNEL_VERSION(5, 5, 0) <= LINUX_VERSION_CODE
 #else
 #include <drm/drmP.h>
 #endif
+#include <drm/drm_print.h>
 #include "evdi_drm_drv.h"
 #include "evdi_params.h"
 #include <linux/shmem_fs.h>
@@ -96,7 +99,7 @@ struct evdi_gem_object *evdi_gem_alloc_object(struct drm_device *dev,
 {
 	struct evdi_gem_object *obj;
 
-	obj = kzalloc(sizeof(*obj), GFP_KERNEL);
+	obj = kzalloc_obj(*obj, GFP_KERNEL);
 	if (obj == NULL)
 		return NULL;
 
@@ -289,6 +292,77 @@ static void evdi_unpin_pages(struct evdi_gem_object *obj)
 		evdi_gem_put_pages(obj);
 	mutex_unlock(&obj->pages_lock);
 }
+# if KERNEL_VERSION(6, 2, 0) <= LINUX_VERSION_CODE || defined(EL8) || defined(EL9)
+#if IS_ENABLED(CONFIG_DRM_TTM_HELPER)
+static bool is_xe_gem_ttm_object_without_vmap(struct dma_buf *dmabuf)
+{
+	struct drm_gem_object *obj;
+
+	if (dmabuf->ops->vmap != drm_gem_dmabuf_vmap || !dmabuf->owner)
+		return false;
+	obj = dmabuf->priv;
+	if (!obj || !obj->funcs)
+		return false;
+
+	if (strncmp("xe", dmabuf->owner->name, min_t(size_t, 2, strlen(dmabuf->owner->name))) != 0)
+		return false;
+
+	return obj->funcs->vmap == NULL;
+}
+
+static int dma_buf_vmap_xe_gem_unlocked(struct dma_buf *dmabuf, struct iosys_map *map)
+{
+	int ret;
+
+	iosys_map_clear(map);
+
+	if (WARN_ON(!dmabuf))
+		return -EINVAL;
+
+	dma_resv_lock(dmabuf->resv, NULL);
+	ret = drm_gem_ttm_vmap(dmabuf->priv, map);
+	dma_resv_unlock(dmabuf->resv);
+
+	return ret;
+}
+
+static void dma_buf_vunmap_xe_gem_unlocked(struct dma_buf *dmabuf, struct iosys_map *map)
+{
+	if (WARN_ON(!dmabuf))
+		return;
+
+	dma_resv_lock(dmabuf->resv, NULL);
+	drm_gem_ttm_vunmap(dmabuf->priv, map);
+	dma_resv_unlock(dmabuf->resv);
+}
+
+#endif
+
+static int evdi_dma_buf_vmap_unlocked(struct dma_buf *dmabuf, struct iosys_map *map)
+{
+	int ret = -EINVAL;
+
+#if IS_ENABLED(CONFIG_DRM_TTM_HELPER)
+	if (is_xe_gem_ttm_object_without_vmap(dmabuf))
+		ret = dma_buf_vmap_xe_gem_unlocked(dmabuf, map);
+#endif
+	if (ret)
+		ret = dma_buf_vmap_unlocked(dmabuf, map);
+	return ret;
+}
+
+static void evdi_dma_buf_vunmap_unlocked(struct dma_buf *dmabuf, struct iosys_map *map)
+{
+#if IS_ENABLED(CONFIG_DRM_TTM_HELPER)
+	if (is_xe_gem_ttm_object_without_vmap(dmabuf))
+		dma_buf_vunmap_xe_gem_unlocked(dmabuf, map);
+	else
+		dma_buf_vunmap_unlocked(dmabuf, map);
+#else
+	dma_buf_vunmap_unlocked(dmabuf, map);
+#endif
+}
+#endif
 
 int evdi_gem_vmap(struct evdi_gem_object *obj)
 {
@@ -304,7 +378,7 @@ int evdi_gem_vmap(struct evdi_gem_object *obj)
 
 #if KERNEL_VERSION(5, 11, 0) <= LINUX_VERSION_CODE || defined(EL8)
 # if KERNEL_VERSION(6, 2, 0) <= LINUX_VERSION_CODE || defined(EL8) || defined(EL9)
-		ret = dma_buf_vmap_unlocked(obj->base.import_attach->dmabuf, &map);
+		ret = evdi_dma_buf_vmap_unlocked(obj->base.import_attach->dmabuf, &map);
 # else
 		ret = dma_buf_vmap(obj->base.import_attach->dmabuf, &map);
 # endif
@@ -342,7 +416,7 @@ void evdi_gem_vunmap(struct evdi_gem_object *obj)
 			iosys_map_set_vaddr(&map, obj->vmapping);
 
 # if KERNEL_VERSION(6, 2, 0) <= LINUX_VERSION_CODE || defined(EL8) || defined(EL9)
-		dma_buf_vunmap_unlocked(obj->base.import_attach->dmabuf, &map);
+		evdi_dma_buf_vunmap_unlocked(obj->base.import_attach->dmabuf, &map);
 # else
 		dma_buf_vunmap(obj->base.import_attach->dmabuf, &map);
 # endif
@@ -402,12 +476,9 @@ int evdi_gem_mmap(struct drm_file *file,
 	struct drm_gem_object *obj;
 	int ret = 0;
 
-	mutex_lock(&dev->struct_mutex);
 	obj = drm_gem_object_lookup(file, handle);
-	if (obj == NULL) {
-		ret = -ENOENT;
-		goto unlock;
-	}
+	if (obj == NULL)
+		return -ENOENT;
 	gobj = to_evdi_bo(obj);
 
 	ret = evdi_pin_pages(gobj);
@@ -429,11 +500,27 @@ int evdi_gem_mmap(struct drm_file *file,
 
  out:
 	drm_gem_object_put(&gobj->base);
- unlock:
-	mutex_unlock(&dev->struct_mutex);
 	return ret;
 }
 
+#if KERNEL_VERSION(5, 8, 0) <= LINUX_VERSION_CODE
+#else
+#define for_each_sgtable_page(sgt, piter, pgoffset)	\
+	for_each_sg_page(sgt->sgl, piter, sgt->orig_nents, pgoffset)
+#endif
+
+static int evdi_get_sg_table_pages_num(struct sg_table *sgt)
+{
+	struct sg_page_iter page_iter;
+	int npages = 0;
+
+	for_each_sgtable_page(sgt, &page_iter, 0) {
+		npages++;
+	}
+
+	return npages;
+}
+
 struct drm_gem_object *
 evdi_prime_import_sg_table(struct drm_device *dev,
 			   struct dma_buf_attachment *attach,
@@ -441,6 +528,7 @@ evdi_prime_import_sg_table(struct drm_device *dev,
 {
 	struct evdi_gem_object *obj;
 	int npages;
+	int npages_sg;
 	bool called_by_mutter;
 
 	called_by_mutter = evdi_was_called_by_mutter();
@@ -450,6 +538,8 @@ evdi_prime_import_sg_table(struct drm_device *dev,
 		return ERR_CAST(obj);
 
 	npages = DIV_ROUND_UP(attach->dmabuf->size, PAGE_SIZE);
+	npages_sg = evdi_get_sg_table_pages_num(sg);
+	npages = npages < npages_sg ? npages_sg : npages_sg;
 	DRM_DEBUG_PRIME("Importing %d pages\n", npages);
 	obj->pages = kvmalloc_array(npages, sizeof(struct page *), GFP_KERNEL);
 	if (!obj->pages) {
